from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING

from tsbot import plugin, query
from tsbot.exceptions import TSException, TSResponseError

from teamspeak_bot.common import CLIENT_LIST_QUERY
from teamspeak_bot.plugins import BasePluginConfig
from teamspeak_bot.utils import cache, find

if TYPE_CHECKING:
    from tsbot import TSBot, TSTask


DEFAULT_AFK_CHANNEL = "AFK"
DEFAULT_IDLE_TIME = 30 * 60


class AFKMoverConfig(BasePluginConfig, total=False):
    afk_channel: str
    idle_time: float


DEFAULT_CONFIG = AFKMoverConfig(
    enabled=True,
)


def is_active(client: dict[str, str], max_idle_time: float):
    return int(client.get("client_idle_time", 0)) < max_idle_time


def in_afk_channel(client: dict[str, str], afk_channel_id: str):
    return client.get("cid", "") == afk_channel_id


def is_query(client: dict[str, str]):
    return client["client_type"] == "1"


class AFKMover(plugin.TSPlugin):
    CHECK_INTERVAL = 60  # Check every minute

    def __init__(self, config: AFKMoverConfig) -> None:
        self.afk_channel = config.get("afk_channel", DEFAULT_AFK_CHANNEL)
        self.idle_time = config.get("idle_time", DEFAULT_IDLE_TIME) * 1000

        self.afk_channel_id: str = ""
        self.task: TSTask | None = None

    def should_be_moved(self, client: dict[str, str]):
        if is_query(client):
            return False

        if is_active(client, self.idle_time):
            return False

        if in_afk_channel(client, self.afk_channel_id):  # noqa: SIM103
            return False

        return True

    async def afk_mover_task(self, bot: TSBot):
        client_list = await cache.with_cache(
            bot.send, CLIENT_LIST_QUERY, max_ttl=self.CHECK_INTERVAL
        )

        to_be_moved = set(map(lambda c: c["clid"], filter(self.should_be_moved, client_list)))
        if not to_be_moved:
            return

        move_query = (
            query("clientmove")
            .params(cid=self.afk_channel_id)
            .param_block({"clid": clid} for clid in to_be_moved)
        )

        with suppress(TSResponseError):
            await bot.send(move_query)

    @plugin.once("connect")
    async def get_afk_channel(self, bot: TSBot, ctx: None):
        channel_list = await bot.send(query("channellist"))

        channel_id = find.from_iterable(channel_list, self.afk_channel, "channel_name", "cid")
        if not channel_id:
            raise TSException("No AFK Channel Found")

        self.afk_channel_id = channel_id

    @plugin.on("connect")
    async def start_afk_mover(self, bot: TSBot, ctx: None):
        self.task = bot.register_every_task(
            self.CHECK_INTERVAL, self.afk_mover_task, name="AFKMover-Task"
        )

    @plugin.on("disconnect")
    async def stop_afk_mover(self, bot: TSBot, ctx: None):
        if self.task is not None:
            self.task = bot.remove_task(self.task)



from __future__ import annotations

from typing import TYPE_CHECKING

from tsbot import plugin, query
from tsbot.exceptions import TSPermissionError
from tsbot.response import TSResponse

from teamspeak_bot.plugins import BasePluginConfig
from teamspeak_bot.utils import checks, try_

if TYPE_CHECKING:
    from tsbot import TSBot, TSCtx


class AdminConfig(BasePluginConfig, total=False):
    allowed_uids: tuple[str]
    allowed_server_groups: tuple[str]
    strict_server_group_checking: bool


DEFAULT_CONFIG = AdminConfig(
    enabled=False,
)


class AdminPlugin(plugin.TSPlugin):
    def __init__(self, bot: TSBot, config: AdminConfig) -> None:
        self.uids = config.get("allowed_uids")
        self.server_groups = config.get("allowed_server_groups")
        self.strict = config.get("strict_server_group_checking", False)

        owner_check = [self.is_allowed_to_run]

        bot.register_command("eval", self.eval_, hidden=True, raw=True, checks=owner_check)
        bot.register_command("exec", self.exec_, hidden=True, raw=True, checks=owner_check)
        bot.register_command("quit", self.quit_, hidden=True, checks=owner_check)
        bot.register_command("send", self.send, hidden=True, raw=True, checks=owner_check)
        bot.register_command("spam", self.spam, hidden=True, checks=owner_check)
        bot.register_command("nickname", self.change_nickname, hidden=True, checks=owner_check)

    async def is_allowed_to_run(self, bot: TSBot, ctx: TSCtx, *args: str, **kwargs: str) -> None:
        if self.uids and checks.check_uids(self.uids, ctx):
            return

        if self.server_groups and await checks.check_groups(
            bot, ctx, self.server_groups, strict=self.strict
        ):
            return

        raise TSPermissionError("Client not permitted to run this command")

    async def eval_(self, bot: TSBot, ctx: TSCtx, eval_str: str) -> None:
        response = try_.or_call(
            eval,
            eval_str,
            globals(),
            locals(),
            on_error=lambda e: type(e).__name__,
        )
        await bot.respond(ctx, response)

    async def exec_(self, bot: TSBot, ctx: TSCtx, exec_str: str) -> None:
        exec(exec_str, globals(), locals())

    async def quit_(self, bot: TSBot, ctx: TSCtx) -> None:
        bot.close()

    async def send(self, bot: TSBot, ctx: TSCtx, raw_command: str) -> None:
        response = await try_.async_or_call(
            bot.send_raw(raw_command), lambda e: f"{type(e).__name__}: {e}"
        )

        await bot.respond(ctx, repr(response) if isinstance(response, TSResponse) else response)

    async def spam(self, bot: TSBot, ctx: TSCtx, amount: str, line: str) -> None:
        for _ in range(int(amount)):
            await bot.respond(ctx, line)

    async def change_nickname(self, bot: TSBot, ctx: TSCtx, nickname: str | None = None) -> None:
        if not nickname:
            resp = await bot.send_raw("whoami")
            nickname = resp.first["client_login_name"]

        await bot.send(query("clientupdate").params(client_nickname=nickname))
        await bot.respond(ctx, f"Nickname changed to {nickname!r}")



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import tsformatter as tsf
from tsbot import plugin

from teamspeak_bot.plugins import BasePluginConfig

if TYPE_CHECKING:
    from tsbot import TSBot, TSCtx


def create_error_prefix(error_name: str):
    return tsf.bold(str.join("", ("[", tsf.color("#f9655d", error_name), "]")))


DEFAULT_MESSAGES = {
    "invalid_invocation_message": f"{create_error_prefix('Argument error')}: {'{exception}'}\nUse '!help {'{command}'}' to see usage.",
    "permission_error_message": f"{create_error_prefix('Permission error')}: {'{exception}'}",
    "permission_error_log_message": "{invokername!r} ({invokeruid}) tried to use command {command!r} with args {raw_args!r}",
    "command_error_message": f"{create_error_prefix('Command error')}: {'{exception}'}",
}


class ErrorEventsConfig(BasePluginConfig, total=False):
    invalid_invocation_message: str
    permission_error_message: str
    permission_error_log_message: str
    command_error_message: str


DEFAULT_CONFIG = ErrorEventsConfig(
    enabled=True,
)


class ErrorEventsPlugin(plugin.TSPlugin):
    def __init__(
        self,
        logger: logging.Logger,
        config: ErrorEventsConfig,
    ) -> None:
        self.logger = logger

        self.invalid_invocation_message = config.get(
            "invalid_invocation_message", DEFAULT_MESSAGES["invalid_invocation_message"]
        )
        self.permission_error_message = config.get(
            "permission_error_message", DEFAULT_MESSAGES["permission_error_message"]
        )
        self.permission_error_log_message = config.get(
            "permission_error_log_message", DEFAULT_MESSAGES["permission_error_log_message"]
        )
        self.command_error_message = config.get(
            "command_error_message", DEFAULT_MESSAGES["command_error_message"]
        )

    @plugin.on("parameter_error")
    async def handle_parameter_error(self, bot: TSBot, ctx: TSCtx):
        await bot.respond(ctx, self.invalid_invocation_message.format(**ctx))

    @plugin.on("command_error")
    async def handle_command_error(self, bot: TSBot, ctx: TSCtx):
        await bot.respond(ctx, self.command_error_message.format(**ctx))

    @plugin.on("permission_error")
    async def handle_permission_error(self, bot: TSBot, ctx: TSCtx):
        self.logger.warning(self.permission_error_log_message.format(**ctx))
        await bot.respond(ctx, self.permission_error_message.format(**ctx))



from __future__ import annotations

import operator
from collections.abc import Iterable, Mapping
from typing import overload


@overload
def from_iterable[K: str, V: str](
    search_list: Iterable[Mapping[K, V]],
    search_str: str,
    search_attr: K,
    result_attr: None = None,
    *,
    strict: bool = False,
) -> Mapping[K, V] | None: ...


@overload
def from_iterable[K: str, V: str](
    search_list: Iterable[Mapping[K, V]],
    search_str: str,
    search_attr: K,
    result_attr: K,
    *,
    strict: bool = False,
) -> V | None: ...


def from_iterable[K: str, V: str](
    search_list: Iterable[Mapping[K, V]],
    search_str: str,
    search_attr: K,
    result_attr: K | None = None,
    *,
    strict: bool = False,
) -> V | Mapping[K, V] | None:
    op = operator.eq if strict else operator.contains

    for item in search_list:
        if op(search_str, item[search_attr]):
            return item[result_attr] if result_attr else item

    return None



from __future__ import annotations

import importlib
import importlib.util
import sys
from typing import Final, Literal, NotRequired, Required, TypedDict, cast

from result import Err, Ok, Result

from teamspeak_bot.plugins import (
    admin,
    afk_mover,
    banned_names,
    error_events,
    fun,
    greeter,
    jokes,
    notify,
)
from teamspeak_bot.utils import merge


class LoggingConfig(TypedDict, total=False):
    console_format: str
    file_format: str


class PluginsConfig(TypedDict, total=False):
    admin: admin.AdminConfig
    afk_mover: afk_mover.AFKMoverConfig
    banned_names: banned_names.BannedNamesConfig
    error_events: error_events.ErrorEventsConfig
    fun: fun.FunConfig
    greeter: greeter.GreeterConfig
    jokes: jokes.JokesPluginConfig
    notify: notify.NotifyConfig


class BotConfig(TypedDict):
    username: Required[str]
    password: Required[str]
    address: Required[str]
    port: NotRequired[int]

    protocol: NotRequired[Literal["ssh", "raw"]]
    server_id: NotRequired[int]
    nickname: NotRequired[str | None]
    invoker: NotRequired[str]
    connection_retries: NotRequired[int]
    connection_retry_timeout: NotRequired[float]
    ratelimited: NotRequired[bool]
    ratelimit_calls: NotRequired[int]
    ratelimit_period: NotRequired[float]
    query_timeout: NotRequired[float]

    plugins: NotRequired[PluginsConfig]

    logging: NotRequired[LoggingConfig]


DEFAULT_PLUGINS_CONFIG: Final[PluginsConfig] = {
    "admin": admin.DEFAULT_CONFIG,
    "afk_mover": afk_mover.DEFAULT_CONFIG,
    "banned_names": banned_names.DEFAULT_CONFIG,
    "error_events": error_events.DEFAULT_CONFIG,
    "fun": fun.DEFAULT_CONFIG,
    "greeter": greeter.DEFAULT_CONFIG,
    "jokes": jokes.DEFAULT_CONFIG,
    "notify": notify.DEFAULT_CONFIG,
}

DEFAULT_CONFIG: BotConfig = {
    "plugins": DEFAULT_PLUGINS_CONFIG,
    "logging": {
        "console_format": "[%(name)s][%(levelname)s] %(message)s",
        "file_format": "%(asctime)s | %(name)-10s | %(levelname)-8s | %(message)s",
    },
}  # type: ignore


def get_config(config_path: str) -> Result[BotConfig, str]:
    spec = importlib.util.spec_from_file_location("config", config_path)
    if spec is None or spec.loader is None:
        return Err("Unable to find 'config' module")

    module = importlib.util.module_from_spec(spec)
    sys.modules["config"] = module
    spec.loader.exec_module(module)

    imported_config = cast(BotConfig | None, getattr(module, "CONFIG", None))
    if imported_config is None:
        return Err("Config module doesn't have CONFIG attribute")

    default_config = DEFAULT_CONFIG.copy()
    return Ok(merge.deep_merge(imported_config, default_config))



from __future__ import annotations

import operator
from collections.abc import Sequence
from typing import TYPE_CHECKING

from teamspeak_bot.utils import get

if TYPE_CHECKING:
    from tsbot import TSBot, TSCtx


def has_group(
    groups: Sequence[str], client_groups: Sequence[dict[str, str]], *, strict: bool = False
) -> bool:
    op = operator.eq if strict else operator.contains
    return any(op(g, cg["name"]) for g in groups for cg in client_groups)


async def check_groups(
    bot: TSBot, ctx: TSCtx, groups: Sequence[str], *, strict: bool = False
) -> bool:
    return has_group(groups, await get.client_server_groups(bot, ctx), strict=strict)


def check_uids(uids: Sequence[str], ctx: TSCtx) -> bool:
    return ctx.get("invokeruid") in uids



from __future__ import annotations

from typing import TYPE_CHECKING

from tsbot import query

if TYPE_CHECKING:
    from tsbot import TSBot, TSCtx


async def client_server_groups(bot: TSBot, ctx: TSCtx) -> tuple[dict[str, str], ...]:
    ids = await bot.send(query("clientgetdbidfromuid").params(cluid=ctx["invokeruid"]))
    groups = await bot.send(query("servergroupsbyclientid").params(cldbid=ids.first["cldbid"]))

    return groups.data



import atexit
import logging
import logging.handlers
import queue
import sys


def calculate_verbosity(level: int):
    if level >= 2:
        return logging.DEBUG
    if level >= 1:
        return logging.INFO
    return logging.WARNING


def create_console_handler(verbosity: int, console_format: str):
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(calculate_verbosity(verbosity))
    handler.setFormatter(logging.Formatter(console_format))
    return handler


def create_file_handler(log_filepath: str, file_format: str):
    handler = logging.handlers.RotatingFileHandler(
        log_filepath, maxBytes=1024 * 1024, backupCount=1
    )
    handler.setLevel(logging.WARNING)
    handler.setFormatter(logging.Formatter(file_format))
    return handler


def create_queue_handler(queue: queue.Queue[logging.LogRecord]):
    handler = logging.handlers.QueueHandler(queue)
    return handler


def create_queue_listener(queue: queue.Queue[logging.LogRecord], *handlers: logging.Handler):
    handler = logging.handlers.QueueListener(queue, *handlers, respect_handler_level=True)
    handler.start()
    atexit.register(handler.stop)
    return handler


def setup_logger(verbosity: int, log_filepath: str, console_format: str, file_format: str):
    console_handler = create_console_handler(verbosity, console_format)
    file_handler = create_file_handler(log_filepath, file_format)

    log_queue: queue.Queue[logging.LogRecord] = queue.Queue()
    queue_handler = create_queue_handler(log_queue)

    create_queue_listener(log_queue, console_handler, file_handler)

    root = logging.getLogger()
    root.setLevel(logging.NOTSET)
    root.addHandler(queue_handler)



from __future__ import annotations

import random
from typing import TYPE_CHECKING

from tsbot import plugin
from tsbot.exceptions import TSCommandError

from teamspeak_bot.plugins import BasePluginConfig
from teamspeak_bot.utils import try_

if TYPE_CHECKING:
    from tsbot import TSBot, TSCtx


class FunConfig(BasePluginConfig): ...


DEFAULT_CONFIG = FunConfig(
    enabled=True,
)


class FunPlugin(plugin.TSPlugin):
    @plugin.command("coinflip", help_text="Heads or tails")
    async def coin_flip(self, bot: TSBot, ctx: TSCtx):
        await bot.respond(ctx, random.choice(("Heads!", "Tails!")))

    @plugin.command("roll", help_text="Roll a die")
    async def roll_a_die(self, bot: TSBot, ctx: TSCtx, sides: str = "6"):
        dice = try_.or_none(int, sides)

        if dice is None or dice <= 0:
            raise TSCommandError("Invalid dice")

        await bot.respond(ctx, f"{random.randint(1, dice)}!")



from __future__ import annotations

import asyncio
from collections.abc import Callable
from contextlib import suppress
from typing import TYPE_CHECKING

from tsbot import plugin, query
from tsbot.exceptions import TSResponseError

from teamspeak_bot.common import CLIENT_LIST_QUERY
from teamspeak_bot.plugins import BasePluginConfig
from teamspeak_bot.utils import cache

if TYPE_CHECKING:
    from tsbot import TSBot, TSCtx, TSTask


REASON_KICK_SERVER = 5
DEFAULT_MESSAGE = "Nickname banned"
DEFAULT_CHECK_PERIOD = 30


class BannedNamesConfig(BasePluginConfig, total=False):
    banned_names: tuple[str, ...]
    is_banned_name: Callable[[str], bool]
    message: str
    check_period: float


DEFAULT_CONFIG = BannedNamesConfig(
    enabled=True,
    banned_names=("TeamSpeakUser",),
)


class BannedNamesPlugin(plugin.TSPlugin):
    KICK_QUERY = query("clientkick").params(reasonid=REASON_KICK_SERVER)

    def __init__(self, bot: TSBot, config: BannedNamesConfig) -> None:
        self.message = config.get("message", DEFAULT_MESSAGE)
        self.banned_names = config.get("banned_names")
        self.is_banned_name = config.get("is_banned_name")
        self.check_period = config.get("check_period", DEFAULT_CHECK_PERIOD)

        self.check_task: TSTask | None = None

        if self.banned_names is None and self.is_banned_name is None:
            raise RuntimeError("Either 'banned_names' or 'is_banned_name' has to be declared")

        bot.register_event_handler("cliententerview", self.check_for_banned_names_on_enter)
        bot.register_event_handler("connect", self.start_check_task)
        bot.register_event_handler("disconnect", self.cancel_check_task)

    async def start_check_task(self, bot: TSBot, ctx: None):
        self.check_task = bot.register_every_task(
            self.check_period, self.check_for_banned_names_periodically
        )

    async def cancel_check_task(self, bot: TSBot, ctx: None):
        if self.check_task:
            self.check_task = bot.remove_task(self.check_task)

    def check_client_nickname(self, nickname: str) -> bool:
        return bool(
            self.banned_names is not None
            and nickname in self.banned_names
            or self.is_banned_name is not None
            and self.is_banned_name(nickname)
        )

    async def check_for_banned_names_on_enter(self, bot: TSBot, ctx: TSCtx):
        if self.check_client_nickname(ctx["client_nickname"]):
            await self.kick_client(bot, ctx["clid"])

    async def check_for_banned_names_periodically(self, bot: TSBot):
        client_list = await cache.with_cache(bot.send, CLIENT_LIST_QUERY, max_ttl=0)

        kick_coros = tuple(
            self.kick_client(bot, client["clid"])
            for client in client_list
            if self.check_client_nickname(client["client_nickname"])
        )

        if not kick_coros:
            return

        await asyncio.gather(*kick_coros)

    async def kick_client(self, bot: TSBot, clid: str):
        with suppress(TSResponseError):
            await bot.send(self.KICK_QUERY.params(clid=clid, reasonmsg=self.message))



from __future__ import annotations

import argparse
import asyncio
import logging
import signal
from contextlib import suppress

from tsbot import TSBot, TSCtx

from teamspeak_bot.config import get_config
from teamspeak_bot.logging import setup_logger
from teamspeak_bot.plugins import (
    admin,
    afk_mover,
    banned_names,
    error_events,
    fun,
    greeter,
    jokes,
    notify,
)

logger = logging.getLogger(__package__)


async def open_console(bot: TSBot, ctx: TSCtx):
    """
    Responds to the client with a dm.

    The bot accepts commands in private chats so you can invoke commands anywhere.
    Plus you don't have to use the invoker on commands inside private chats.
    """
    await bot.respond_to_client(ctx, "Console opened")


async def async_main(bot: TSBot):
    with suppress(NotImplementedError):
        asyncio.get_running_loop().add_signal_handler(signal.SIGINT, bot.close)

    await bot.run()


def main():
    parser = argparse.ArgumentParser(prog="TeamSpeak Bot", description="TeamSpeak Server Query Bot")

    parser.add_argument("-c", "--config", default="config.py", help="Path to a config file")
    parser.add_argument("-v", "--verbose", action="count", default=0, help="Verbosity level")
    parser.add_argument("-l", "--logfile", default="log.txt", help="Path to log file")

    args = parser.parse_args()
    config = get_config(args.config).unwrap_or_raise(RuntimeError)

    logging_config = config["logging"]  # type: ignore
    formats = logging_config["console_format"], logging_config["file_format"]  # type: ignore

    setup_logger(args.verbose, args.logfile, *formats)

    extra_kwargs = {
        k: v
        for k in (
            "port",
            "protocol",
            "server_id",
            "nickname",
            "invoker",
            "connection_retries",
            "connection_retry_timeout",
            "ratelimited",
            "ratelimit_calls",
            "ratelimit_period",
            "query_timeout",
        )
        if (v := config.get(k)) is not None
    }

    bot = TSBot(
        username=config["username"],
        password=config["password"],
        address=config["address"],
        **extra_kwargs,
    )

    bot.register_command(
        "console",
        open_console,
        help_text="Sends the client a private message. Clients can use it to invoke commands anywhere",
    )

    if plugins_config := config.get("plugins"):
        if (admin_config := plugins_config.get("admin")) and admin_config["enabled"]:
            bot.load_plugin(admin.AdminPlugin(bot, admin_config))

        if (afk_config := plugins_config.get("afk_mover")) and afk_config["enabled"]:
            bot.load_plugin(afk_mover.AFKMover(afk_config))

        if (banned_names_cfg := plugins_config.get("banned_names")) and banned_names_cfg["enabled"]:
            bot.load_plugin(banned_names.BannedNamesPlugin(bot, banned_names_cfg))

        if (error_events_cfg := plugins_config.get("error_events")) and error_events_cfg["enabled"]:
            bot.load_plugin(error_events.ErrorEventsPlugin(logger, error_events_cfg))

        if (fun_config := plugins_config.get("fun")) and fun_config["enabled"]:
            bot.load_plugin(fun.FunPlugin())

        if (greeter_config := plugins_config.get("greeter")) and greeter_config["enabled"]:
            bot.load_plugin(greeter.GreeterPlugin(greeter_config))

        if (jokes_config := plugins_config.get("jokes")) and jokes_config["enabled"]:
            bot.load_plugin(jokes.JokesPlugin())

        if (notify_config := plugins_config.get("notify")) and notify_config["enabled"]:
            bot.load_plugin(notify.NotifyPlugin(notify_config))

    raise SystemExit(asyncio.run(async_main(bot)))


if __name__ == "__main__":
    main()



from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING

from tsbot import plugin, query
from tsbot.exceptions import TSException, TSResponseError

from teamspeak_bot.plugins import BasePluginConfig

if TYPE_CHECKING:
    from tsbot import TSBot, TSCtx


DEFAULT_MESSAGE = "Welcome to server, {client_nickname}"


class GreeterConfig(BasePluginConfig, total=False):
    message: str


DEFAULT_CONFIG = GreeterConfig(
    enabled=True,
)


class GreeterPlugin(plugin.TSPlugin):
    POKE_QUERY = query("clientpoke")

    def __init__(self, config: GreeterConfig) -> None:
        self.message = config.get("message", DEFAULT_MESSAGE)

        self.guest_id: str = ""

    @plugin.once("connect")
    async def get_guest_id(self, bot: TSBot, ctx: None):
        server_groups = await bot.send(query("servergrouplist"))

        guest_id = None
        for group in server_groups:
            if group["name"] == "Guest" and group["type"] == "1":
                guest_id = group["sgid"]

        if not guest_id:
            raise TSException("Failed to find 'Guest' server id")

        self.guest_id = guest_id

    @plugin.on("cliententerview")
    async def handle_client_enter(self, bot: TSBot, ctx: TSCtx):
        # Only greet users who have no server groups
        if ctx["client_servergroups"] != self.guest_id:
            return

        poke_query = self.POKE_QUERY.params(clid=ctx["clid"], msg=self.message.format(**ctx))
        with suppress(TSResponseError):
            await bot.send(poke_query)
